;;; Directory Local Variables
;;; For more information see (info "(emacs) Directory Variables")

((nil . ((tab-width . 8)
         (fill-column . 78)
         (sentence-end-double-space . t)))
 (lisp-mode . ((indent-tabs-mode . nil))))
